@@ -105,7 +105,10 @@ rand_strings <- function(n) {
 
 nested_tibble <- function() {
   sub_tibble <- function(nr = 600, nc = 4) {
-    z <- lapply(1:nc, function(i) rand_strings(nr)) %>% bind_cols %>% setNames(make.unique(paste0(sample(letters, nc), rand_strings(nc)))) %>% as_tibble
+    z <- lapply(1:nc, function(i) rand_strings(nr)) %>%
+      setNames(make.unique(paste0(sample(letters, nc), rand_strings(nc)))) %>%
+      bind_cols %>%
+      as_tibble
   }
   tibble(
     col1 = rand_strings(100),
@@ -477,7 +480,7 @@ for (q in 1:reps) {
 
   time <- vector("numeric", length = 3)
   for (i in 1:3) {
-    x1 <- suppressMessages(nested_tibble())
+    x1 <- nested_tibble()
     time[i] <- Sys.time()
     qsave_rand(x1, file = myfile)
     z <- qread_rand(file = myfile)
